"""S3 file source module for retrieving objects from S3 buckets.

This module provides:
- S3FileSource: A source implementation for listing and downloading files from S3.
"""
import logging
from pathlib import Path
from typing import Any
from tgedr_dataops.commons.s3_connector import S3Connector

from tgedr_dataops_abs.source import Source, SourceException
from tgedr_dataops.commons.utils_fs import remove_s3_protocol, resolve_s3_protocol


logger = logging.getLogger(__name__)


class S3FileSource(Source, S3Connector):
    """class used to retrieve objects/files from s3 bucket to local fs location."""

    CONTEXT_KEY_SOURCE = "source"
    CONTEXT_KEY_TARGET = "target"
    CONTEXT_KEY_FILES = "files"
    CONTEXT_KEY_SUFFIX = "suffix"

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        """Initialize the S3FileSource.

        Parameters
        ----------
        config : dict[str, Any], optional
            Configuration dictionary containing optional AWS credentials.
        """
        Source.__init__(self, config=config)
        S3Connector.__init__(self)

    def list(self, context: dict[str, Any] | None = None) -> list[str]:
        """List files in the S3 bucket.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' (S3 URL) and optionally 'suffix' for filtering.

        Returns
        -------
        list[str]
            List of S3 object URLs in the source bucket/prefix.

        Raises
        ------
        SourceException
            If source context is missing.
        """
        logger.info(f"[list|in] ({context})")

        result: list[str] = []
        if self.CONTEXT_KEY_SOURCE not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_SOURCE}")

        s3_protocol: str = resolve_s3_protocol(context[self.CONTEXT_KEY_SOURCE])
        protocol = "" if s3_protocol is None else s3_protocol

        path = remove_s3_protocol(context[self.CONTEXT_KEY_SOURCE])
        path_elements = path.split("/")
        bucket = path_elements[0]
        key = "/".join(path_elements[1:])

        objs = self._client.list_objects_v2(Bucket=bucket, Prefix=key)
        result = [
            (protocol + bucket + "/" + entry["Key"]) for entry in objs["Contents"] if not (entry["Key"]).endswith("/")
        ]

        if self.CONTEXT_KEY_SUFFIX in context:
            suffix: str = context[self.CONTEXT_KEY_SUFFIX]
            result = [f for f in result if f.endswith(suffix)]

        logger.debug(f"[list|out] => {result}")
        logger.info(f"[list|out] => result len: {len(result)}")
        return result

    def get(self, context: dict[str, Any] | None = None) -> Any:
        """Download a file from S3 to local filesystem.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' (S3 URL) and 'target' (local path).

        Returns
        -------
        Any
            The local target file path.

        Raises
        ------
        SourceException
            If source or target context is missing.
        """
        logger.info(f"[get|in] ({context})")

        result: list[str] = []
        if self.CONTEXT_KEY_FILES not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_FILES}")
        if self.CONTEXT_KEY_TARGET not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_TARGET}")

        files = context[self.CONTEXT_KEY_FILES]
        target = context[self.CONTEXT_KEY_TARGET]

        target_is_dir: bool = Path(target).is_dir()
        target = target.rstrip("/") if target.endswith("/") else target

        for file in files:
            path_elements = remove_s3_protocol(file).split("/")
            bucket = path_elements[0]
            key = "/".join(path_elements[1:])
            filename = path_elements[-1]
            local_file = str(Path(target) / filename) if target_is_dir else target

            # assure we have that path there
            local_folder = Path(local_file).parent
            if not local_folder.is_dir():
                local_folder.mkdir(parents=True)

            logger.info(f"[get] bucket: {bucket}   key: {key}   file: {file}   local_file: {local_file}")
            self._client.download_file(Bucket=bucket, Key=key, Filename=local_file)
            result.append(local_file)

        logger.info(f"[get|out] => {result}")
        return result



"""Local filesystem source implementation for file operations.

This module provides:
- LocalFsFileSource: A source class for retrieving and listing local files.
"""
import logging
import shutil
from pathlib import Path
from typing import Any

from tgedr_dataops_abs.source import Source, SourceException


logger = logging.getLogger(__name__)


class LocalFsFileSource(Source):
    """source class used to retrieve local objects/files to a another local fs location."""

    CONTEXT_KEY_SOURCE = "source"
    CONTEXT_KEY_TARGET = "target"
    CONTEXT_KEY_SUFFIX = "suffix"
    CONTEXT_KEY_FILES = "files"

    def list(self, context: dict[str, Any] | None = None) -> list[str]:
        """List files in the local filesystem directory.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' path.

        Returns
        -------
        list[str]
            List of file paths in the source directory.

        Raises
        ------
        SourceException
            If source context is missing or if source path is not a directory.
        """
        logger.info(f"[list|in] ({context})")
        result: list[str] = []
        if self.CONTEXT_KEY_SOURCE not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_SOURCE}")

        source = context[self.CONTEXT_KEY_SOURCE]
        source_path = Path(source)
        if source_path.is_dir():
            suffix = None
            if self.CONTEXT_KEY_SUFFIX in context:
                suffix = context[self.CONTEXT_KEY_SUFFIX]
                result: list[str] = [str(file) for file in source_path.iterdir() if file.name.endswith(suffix)]
            else:
                result: list[str] = [str(file) for file in source_path.iterdir()]
        elif source_path.is_file():
            result: list[str] = [source]

        logger.debug(f"[list|out] => {result}")
        logger.info(f"[list|out] => result len: {len(result)}")
        return result

    def get(self, context: dict[str, Any] | None = None) -> Any:
        """Retrieve file(s) from local filesystem.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' path.

        Returns
        -------
        Any
            List of file paths if source is a directory, or single file path if source is a file.

        Raises
        ------
        SourceException
            If source context is missing.
        """
        logger.info(f"[get|in] ({context})")

        if self.CONTEXT_KEY_FILES not in context or self.CONTEXT_KEY_TARGET not in context:
            raise SourceException(f"{self.CONTEXT_KEY_FILES} and {self.CONTEXT_KEY_TARGET} must be provided in config")
        files = context[self.CONTEXT_KEY_FILES]
        target = context[self.CONTEXT_KEY_TARGET]

        if "list" != type(files).__name__:
            if "str" == type(files).__name__:
                files = [files]
            else:
                raise SourceException("files argument must be a list of strings or a string")

        target_is_dir: bool = False
        if Path(target).is_dir():
            target_is_dir = True

        result: list[str] = []

        for file in files:
            basename = Path(file).name
            new_file = str(Path(target) / basename) if target_is_dir else target
            shutil.copy(file, new_file)
            result.append(new_file)

        logger.info("[get|out] => {result}")
        return result



"""Module for reading pandas DataFrames from S3 sources.

This module provides:
- PdDfS3Source: class for reading CSV and Excel files from S3 into pandas DataFrames.
"""
from io import StringIO
import logging
from typing import Any, ClassVar
import pandas as pd

from tgedr_dataops.commons.utils_fs import process_s3_url
from tgedr_dataops.source.abstract_s3_file_source import AbstractS3FileSource
from tgedr_dataops_abs.source import SourceException

logger = logging.getLogger()


class PdDfS3Source(AbstractS3FileSource):
    """class used to read a pandas dataframe from a csv file in s3."""

    CONTEXT_KEY_FILE_FORMAT = "file_format"
    CONTEXT_KEY_SEPARATOR = "sep"
    CONTEXT_KEY_NO_HEADER = "no_header"
    CONTEXT_KEY_COLUMN_NAMES = "column_names"
    CONTEXT_KEY_SCHEMA_TYPES = "schema_types"
    DEFAULT_FORMAT = "csv"
    FORMATS: ClassVar[list[str]] = ["csv", "xlsx"]
    DEFAULT_SEPARATOR = ","

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        """Initialize the PdDfS3Source with optional configuration.

        Parameters
        ----------
        config : dict[str, Any], optional
            Configuration dictionary for S3 connection, by default None.
        """
        super().__init__(config=config)

    def get(self, context: dict[str, Any] | None = None) -> pd.DataFrame:
        """Retrieve and load a pandas DataFrame from S3.

        Supports CSV and Excel formats.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing:
            - 'source' or CONTEXT_KEY_URL: S3 URL
            - 'file_format': File format (csv or xlsx), defaults to csv
            - 'sep': CSV separator, defaults to comma
            - 'no_header': If present, CSV has no header row
            - 'column_names': List of column names for CSV
            - 'schema_types': Dictionary of column types for CSV

        Returns
        -------
        pd.DataFrame
            The loaded pandas DataFrame.

        Raises
        ------
        SourceException
            If source URL is missing or if file_format is unsupported.
        """
        logger.info(f"[get|in] ({context})")
        result: pd.DataFrame = None

        if self.CONTEXT_KEY_URL not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_URL}")

        _format: str = self.DEFAULT_FORMAT
        if self.CONTEXT_KEY_FILE_FORMAT in context:
            _format = context[self.CONTEXT_KEY_FILE_FORMAT]
            if _format not in self.FORMATS:
                raise SourceException(f"[get] invalid format: {_format}")

        result = self.__read_csv(context=context) if "csv" == _format else self.__read_excel(context=context)

        logger.info(f"[get|out] => {result}")
        return result

    def __read_csv(self, context: dict[str, Any] | None = None) -> pd.DataFrame:
        """Read a CSV file from S3 into a pandas DataFrame.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary with S3 URL and CSV reading options.

        Returns
        -------
        pd.DataFrame
            The loaded DataFrame from CSV.
        """
        logger.info(f"[__read_csv|in] ({context})")

        _, bucket, key = process_s3_url(context[self.CONTEXT_KEY_URL])

        obj = self._client.get_object(Bucket=bucket, Key=key)
        data = obj["Body"].read().decode("utf-8")

        header = 0 if self.CONTEXT_KEY_NO_HEADER not in context else None
        names = context.get(self.CONTEXT_KEY_COLUMN_NAMES, None)
        dtype = context.get(self.CONTEXT_KEY_SCHEMA_TYPES, None)
        sep = context.get(self.CONTEXT_KEY_SEPARATOR, self.DEFAULT_SEPARATOR)

        result: pd.DataFrame = pd.read_csv(StringIO(data), sep=sep, header=header, names=names, dtype=dtype)

        logger.info(f"[__read_csv|out] => {result}")
        return result

    def __read_excel(self, context: dict[str, Any] | None = None) -> pd.DataFrame:
        """Read an Excel file from S3 into a pandas DataFrame.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing S3 URL.

        Returns
        -------
        pd.DataFrame
            The loaded DataFrame from Excel.
        """
        logger.info(f"[__read_excel|in] ({context})")
        src = context[self.CONTEXT_KEY_URL]
        result: pd.DataFrame = pd.read_excel(src, engine="openpyxl")
        logger.info(f"[__read_excel|out] => {result}")
        return result



"""S3 implementation of single partition parquet store."""
import logging
import s3fs
from typing import Any
import pandas as pd
import pyarrow as pa

from tgedr_dataops.store.fs_single_partition_parquet import FsSinglePartitionParquetStore
from tgedr_dataops.commons.utils_fs import remove_s3_protocol


logger = logging.getLogger(__name__)


class S3FsSinglePartitionParquetStore(FsSinglePartitionParquetStore):  # pragma: no cover
    """FsSinglePartitionParquetStore implementation using aws s3 file system."""

    CONFIG_KEY_AWS_ACCESS_KEY_ID: str = "aws_access_key_id"
    CONFIG_KEY_AWS_SECRET_ACCESS_KEY: str = "aws_secret_access_key"  # noqa: S105
    CONFIG_KEY_AWS_SESSION_TOKEN: str = "aws_session_token"  # noqa: S105

    @property
    def fs(self) -> Any:
        """Get the S3 filesystem instance.

        Returns
        -------
        Any
            The s3fs.S3FileSystem instance for S3 operations.
        """
        if self._fs is None:
            if (self._config is not None) and all(
                element in list(self._config.keys())
                for element in [
                    self.CONFIG_KEY_AWS_ACCESS_KEY_ID,
                    self.CONFIG_KEY_AWS_SECRET_ACCESS_KEY,
                    self.CONFIG_KEY_AWS_SESSION_TOKEN,
                ]
            ):
                self._fs = s3fs.S3FileSystem(
                    key=self._config[self.CONFIG_KEY_AWS_ACCESS_KEY_ID],
                    secret=self._config[self.CONFIG_KEY_AWS_SECRET_ACCESS_KEY],
                    token=self._config[self.CONFIG_KEY_AWS_SESSION_TOKEN],
                )
            else:
                self._fs = s3fs.S3FileSystem()
        return self._fs

    def _rmdir(self, key: str) -> None:
        """Remove a directory (prefix) from S3.

        Parameters
        ----------
        key : str
            S3 path/prefix to delete.
        """
        if self.fs.isdir(key):
            self.fs.delete(key, recursive=True)

    def _exists(self, key: str) -> bool:
        """Check if a path exists in S3.

        Parameters
        ----------
        key : str
            S3 path to check for existence.

        Returns
        -------
        bool
            True if path exists in S3, False otherwise.
        """
        return self.fs.get_file_info(key).type.name != "NotFound"

    def get(
        self,
        key: str,
        filter_func: callable | None = None,
        filters: list[tuple[str, str, list[str]]] | None = None,
        schema: pa.Schema = None,
    ) -> pd.DataFrame:
        """Retrieve data from S3 parquet store.

        Parameters
        ----------
        key : str
            S3 path to the parquet data (s3:// protocol will be removed).
        filter_func : callable, optional
            Row filter function.
        filters : list[tuple[str, str, list[str]]], optional
            Partition filters.
        schema : pa.Schema, optional
            PyArrow schema for reading.

        Returns
        -------
        pd.DataFrame
            The loaded DataFrame.
        """
        return super().get(key=remove_s3_protocol(key), filter_func=filter_func, filters=filters, schema=schema)

    def delete(
        self,
        key: str,
        partition_field: str | None = None,
        partition_values: list[str] | None = None,
        kv_dict: dict[str, list[Any]] | None = None,
        schema: pa.Schema = None,
    ) -> None:
        """Delete data from S3 parquet store.

        Parameters
        ----------
        key : str
            S3 path to the parquet data (s3:// protocol will be removed).
        partition_field : str, optional
            Field used for partitioning.
        partition_values : list[str], optional
            List of partition values to delete.
        kv_dict : dict[str, list[Any]], optional
            Dictionary of key-value filters for deletion.
        schema : pa.Schema, optional
            PyArrow schema.
        """
        super().delete(
            key=remove_s3_protocol(key),
            partition_field=partition_field,
            partition_values=partition_values,
            kv_dict=kv_dict,
            schema=schema,
        )

    def save(
        self,
        df: pd.DataFrame,
        key: str,
        partition_field: str | None = None,
        append: bool = False,
        replace_partitions: bool = False,
        schema: Any = None,
    ) -> None:
        """Save DataFrame to S3 parquet store.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to save.
        key : str
            S3 path to save the parquet data (s3:// protocol will be removed).
        partition_field : str, optional
            Field to partition by.
        append : bool, default False
            If True, append to existing data.
        replace_partitions : bool, default False
            If True, replace existing partitions.
        schema : Any, optional
            Schema for the data.
        """
        super().save(
            df=df,
            key=remove_s3_protocol(key),
            partition_field=partition_field,
            append=append,
            replace_partitions=replace_partitions,
            schema=schema,
        )

    def update(
        self,
        df: pd.DataFrame,
        key: str,
        key_fields: list[str],
        partition_field: str | None = None,
        schema: Any = None,
    ) -> None:
        """Update data in S3 parquet store.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame with updated data.
        key : str
            S3 path to the parquet data (s3:// protocol will be removed).
        key_fields : list[str]
            List of fields to use as keys for matching records.
        partition_field : str, optional
            Field used for partitioning.
        schema : Any, optional
            Schema for the data.
        """
        super().update(
            df=df, key=remove_s3_protocol(key), key_fields=key_fields, partition_field=partition_field, schema=schema
        )



"""S3 file extended source module for retrieving objects from S3 with metadata support.

This module provides:
- S3FileExtendedSource: class for retrieving S3 objects/files with metadata extraction
"""
import logging
from typing import Any, ClassVar

from tgedr_dataops.source.s3_file_source import S3FileSource
from tgedr_dataops_abs.source import SourceException
from tgedr_dataops.commons.utils_fs import process_s3_path


logger = logging.getLogger(__name__)


class S3FileExtendedSource(S3FileSource):
    """Class used to retrieve objects/files from S3 bucket to local filesystem."""

    METADATA_KEYS: ClassVar[list[str]] = ["LastModified", "ContentLength", "ETag", "VersionId", "ContentType"]

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        """Initialize the S3FileExtendedSource.

        Parameters
        ----------
        config : dict[str, Any], optional
            Configuration dictionary containing optional AWS credentials.
        """
        super().__init__(config=config)

    def get_metadata(self, context: dict[str, Any] | None = None) -> dict[str, Any]:
        """Retrieve metadata for an S3 object.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' S3 URL.

        Returns
        -------
        dict[str, Any]
            Dictionary containing object metadata including size, last modified time, etc.

        Raises
        ------
        SourceException
            If source context is missing.
        """
        logger.info(f"[get_metadata|in] ({context})")

        result: dict[str, Any] = {}
        if self.CONTEXT_KEY_SOURCE not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_SOURCE}")

        bucket, key = process_s3_path(context[self.CONTEXT_KEY_SOURCE])

        o = self._client.head_object(Bucket=bucket, Key=key)

        for key in list(o.keys()):
            if key in self.METADATA_KEYS:
                if key == "LastModified":
                    result[key] = int(o[key].timestamp())
                else:
                    result[key] = o[key]

        logger.info(f"[get_metadata|out] => result len: {len(result)}")
        return result



"""AWS S3 connector module.

This module provides the S3Connector class for establishing and managing
connections to AWS S3 resources.
"""
import os
import boto3


class S3Connector:
    """utility base class to be extended, providing a connection session with aws s3 resources."""

    def __init__(self) -> None:
        """Initialize the S3Connector with empty session, resource, and client attributes."""
        self.__resource = None
        self.__session = None
        self.__client = None

    @property
    def _session(self) -> boto3.Session:
        """Get or create a boto3 session.

        Returns
        -------
        boto3.Session
            The boto3 session instance, using credentials if configured.
        """
        if self.__session is None:
            if "1" == os.getenv("S3_CONNECTOR_USE_CREDENTIALS", default="0"):
                self.__session = boto3.Session(
                    aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
                    aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
                    aws_session_token=os.environ["AWS_SESSION_TOKEN"],
                    region_name=os.environ["AWS_REGION"],
                )
            else:
                self.__session = boto3.Session()

        return self.__session

    @property
    def _resource(self) -> boto3.resources.base.ServiceResource:
        """Get or create an S3 resource.

        Returns
        -------
        boto3.resources.base.ServiceResource
            The boto3 S3 resource instance.
        """
        if self.__resource is None:
            self.__resource = self._session.resource("s3")
        return self.__resource

    @property
    def _client(self) -> boto3.client:
        """Get or create an S3 client.

        Returns
        -------
        boto3.client
            The boto3 S3 client instance.
        """
        if self.__client is None:
            self.__client = self._session.client("s3")
        return self.__client



"""S3 sink implementation for persisting files to S3.

This module provides:
- S3FileSink: sink class for saving/persisting files to S3 buckets.
"""
import logging
from pathlib import Path
from typing import Any
from tgedr_dataops.commons.s3_connector import S3Connector

from tgedr_dataops_abs.sink import Sink, SinkException
from tgedr_dataops.commons.utils_fs import remove_s3_protocol


logger = logging.getLogger(__name__)


class S3FileSink(Sink, S3Connector):
    """sink class used to save/persist a local object/file to an s3 bucket."""

    CONTEXT_SOURCE_PATH = "source"
    CONTEXT_TARGET_PATH = "target"

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        """Initialize the S3FileSink.

        Parameters
        ----------
        config : dict[str, Any], optional
            Configuration dictionary.
        """
        Sink.__init__(self, config=config)
        S3Connector.__init__(self)

    def put(self, context: dict[str, Any] | None = None) -> Any:
        """Upload a local file to S3 bucket.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' (local file) and 'target' (S3 URL) paths.

        Raises
        ------
        SinkException
            If source or target context is missing, or if source is a directory.
        """
        logger.info(f"[put|in] ({context})")

        if self.CONTEXT_SOURCE_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_SOURCE_PATH}")
        if self.CONTEXT_TARGET_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_TARGET_PATH}")

        source = context[self.CONTEXT_SOURCE_PATH]
        if Path(source).is_dir():
            raise SinkException("source can't be a folder, must be a file")

        target = remove_s3_protocol(context[self.CONTEXT_TARGET_PATH])
        target_elements = target.split("/")
        target_bucket = target_elements[0]
        target_key = "/".join(target_elements[1:])

        if target_key.endswith("/"):
            target_file = Path(source).name
            target_key = target_key + target_file

        logger.info(f"[put] uploading {source} to key: {target_key} in bucket: {target_bucket}")
        self._client.upload_file(Filename=source, Bucket=target_bucket, Key=target_key)

        logger.info("[put|out]")

    def delete(self, context: dict[str, Any] | None = None) -> None:
        """Delete an object from S3 bucket.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'target' (S3 URL) path.

        Raises
        ------
        SinkException
            If target context is missing.
        """
        logger.info(f"[delete|in] ({context})")

        if self.CONTEXT_TARGET_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_TARGET_PATH}")

        target = remove_s3_protocol(context[self.CONTEXT_TARGET_PATH])
        target_elements = target.split("/")
        bucket = target_elements[0]
        key = "/".join(target_elements[1:])

        response = self._client.delete_object(Bucket=bucket, Key=key)
        logger.info(f"[delete] response: {response}")
        logger.info("[delete|out]")



"""Local filesystem implementation for single partition Parquet storage.

This module provides LocalFsSinglePartitionParquetStore, which implements
single partition Parquet storage using the local file system.
"""
import logging
from typing import Any
from pyarrow import fs

from tgedr_dataops.store.fs_single_partition_parquet import FsSinglePartitionParquetStore


logger = logging.getLogger(__name__)


class LocalFsSinglePartitionParquetStore(FsSinglePartitionParquetStore):
    """FsSinglePartitionParquetStore implementation using local file system."""

    @property
    def fs(self) -> Any:
        """Get the PyArrow local filesystem instance.

        Returns
        -------
        Any
            The filesystem instance for local file operations.
        """
        if self._fs is None:
            self._fs = fs.LocalFileSystem()
        return self._fs

    def _rmdir(self, key: str) -> None:
        """Remove a directory recursively.

        Parameters
        ----------
        key : str
            Path to the directory to delete.
        """
        if self.fs.get_file_info(key).type.name == "Directory":
            self.fs.delete_dir(key)

    def _exists(self, key: str) -> bool:
        """Check if a path exists on the local filesystem.

        Parameters
        ----------
        key : str
            Path to check for existence.

        Returns
        -------
        bool
            True if path exists, False otherwise.
        """
        return self.fs.get_file_info(key).type.name != "NotFound"



"""Abstract base class for S3 file sources.

This module provides:
- AbstractS3FileSource: abstract class for reading file sources from S3.
"""
from abc import ABC
import logging
from typing import Any

from tgedr_dataops.commons.s3_connector import S3Connector
from tgedr_dataops.commons.utils_fs import process_s3_url
from tgedr_dataops_abs.source import Source, SourceException


logger = logging.getLogger()


class AbstractS3FileSource(Source, S3Connector, ABC):
    """abstract class used to read file sources from s3."""

    CONTEXT_KEY_URL = "url"
    CONTEXT_KEY_SUFFIX = "suffix"

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        """Initialize the S3 file source.

        Parameters
        ----------
        config : dict[str, Any], optional
            Configuration dictionary for the source.
        """
        Source.__init__(self, config=config)
        S3Connector.__init__(self)

    def list(self, context: dict[str, Any] | None = None) -> list[str]:
        """List objects in the S3 bucket.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' S3 URL.

        Returns
        -------
        list[str]
            List of S3 object keys in the source bucket/prefix.

        Raises
        ------
        SourceException
            If source context is missing.
        """
        logger.info(f"[list|in] ({context})")

        result: list[str] = []
        if self.CONTEXT_KEY_URL not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_URL}")

        protocol, bucket, key = process_s3_url(context[self.CONTEXT_KEY_URL])

        objs = self._client.list_objects_v2(Bucket=bucket, Prefix=key)
        result = [
            (protocol + bucket + "/" + entry["Key"]) for entry in objs["Contents"] if not (entry["Key"]).endswith("/")
        ]

        if self.CONTEXT_KEY_SUFFIX in context:
            suffix: str = context[self.CONTEXT_KEY_SUFFIX]
            result = [f for f in result if f.endswith(suffix)]

        logger.debug(f"[list|out] => {result}")
        logger.info(f"[list|out] => result len: {len(result)}")
        return result



"""Pandas DataFrame validation implementation module.

This module provides the Pandas-specific implementation of Great Expectations validation.
"""
from great_expectations.execution_engine import ExecutionEngine
from great_expectations.execution_engine import PandasExecutionEngine
from tgedr_dataops_abs.great_expectations_validation import GreatExpectationsValidation


class PandasValidation(GreatExpectationsValidation):
    """Pandas DataFrame validation implementation."""

    def _get_execution_engine(self, batch_data_dict: dict) -> ExecutionEngine:
        """Get the execution engine used by the validation implementation.

        Returns
        -------
        ExecutionEngine
            The execution engine instance.
        """
        return PandasExecutionEngine(batch_data_dict=batch_data_dict)



"""Local filesystem sink implementation for persisting files.

This module provides:
- LocalFsFileSink: sink class for saving/persisting files to local filesystem.
"""

import logging
import shutil
from pathlib import Path
from typing import Any

from tgedr_dataops_abs.sink import Sink, SinkException


logger = logging.getLogger(__name__)


class LocalFsFileSink(Sink):
    """sink class used to save/persist an object/file to a local fs location."""

    CONTEXT_SOURCE_PATH = "source"
    CONTEXT_TARGET_PATH = "target"

    def put(self, context: dict[str, Any] | None = None) -> Any:
        """Copy a file from source to target on local filesystem.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' and 'target' paths.

        Raises
        ------
        SinkException
            If source or target context is missing.
        """
        logger.info(f"[put|in] ({context})")

        if self.CONTEXT_SOURCE_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_SOURCE_PATH}")
        if self.CONTEXT_TARGET_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_TARGET_PATH}")

        source = context[self.CONTEXT_SOURCE_PATH]
        target = context[self.CONTEXT_TARGET_PATH]

        shutil.copy(source, target)
        logger.info("[put|out]")

    def delete(self, context: dict[str, Any] | None = None) -> None:
        """Delete a file or directory from local filesystem.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'target' path.

        Raises
        ------
        SinkException
            If target context is missing or target is neither file nor directory.
        """
        logger.info(f"[delete|in] ({context})")

        if self.CONTEXT_TARGET_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_TARGET_PATH}")

        target = Path(context[self.CONTEXT_TARGET_PATH])

        if target.is_file():
            target.unlink()
        elif target.is_dir():
            shutil.rmtree(target)
        else:
            raise SinkException(f"[delete] is it a dir or a folder? {target}")

        logger.info("[delete|out]")



"""Filesystem-based single partition Parquet store implementation.

This module provides an abstract base class for storing and retrieving data
in Parquet format with optional single partition support across different
filesystem implementations (local, S3, etc.).
"""
from abc import ABC, abstractmethod
import logging
from pathlib import Path
from typing import Any
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.compute as pc

from tgedr_dataops_abs.store import Store, StoreException


logger = logging.getLogger(__name__)


def pandas_mapper(arrow_type: pa.DataType) -> pd.api.extensions.ExtensionDtype | None:
    """Map PyArrow types to pandas nullable types.

    Parameters
    ----------
    arrow_type : pa.DataType
        PyArrow data type to map.

    Returns
    -------
    pd.api.extensions.ExtensionDtype or None
        Corresponding pandas nullable dtype, or None for default behavior.
    """
    if pa.types.is_int64(arrow_type):
        return pd.Int64Dtype()
    if pa.types.is_float64(arrow_type):
        return pd.Float64Dtype()
    if pa.types.is_string(arrow_type):
        return pd.StringDtype()   # pragma: no cover
    # suggest default behavior
    return None


class FsSinglePartitionParquetStore(Store, ABC):
    """Abstract store implementation for Parquet files with optional single partition.

    This class provides persistence on Parquet files with an optional single partition,
    regardless of the underlying filesystem location (local, S3, etc.).
    """

    @property
    @abstractmethod
    def fs(self) -> Any:
        """Abstract property providing a filesystem implementation.

        Returns
        -------
        Any
            Filesystem implementation (e.g., LocalFileSystem, S3FileSystem).
        """
        raise NotImplementedError

    @abstractmethod
    def _rmdir(self, key: str) -> None:
        """Remove a directory.

        Parameters
        ----------
        key : str
            Directory path to remove.
        """
        raise NotImplementedError

    @abstractmethod
    def _exists(self, key: str) -> bool:
        """Check if a path exists.

        Parameters
        ----------
        key : str
            Path to check for existence.

        Returns
        -------
        bool
            True if path exists, False otherwise.
        """
        raise NotImplementedError

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        """Initialize the FsSinglePartitionParquetStore.

        Parameters
        ----------
        config : dict[str, Any], optional
            Configuration dictionary.
        """
        Store.__init__(self, config)
        self._fs = None

    def get(
        self,
        key: str,
        filter_func: Any | None = None,
        filters: list[tuple[str, str, list[str]]] | None = None,
        schema: pa.Schema | None = None,
    ) -> pd.DataFrame:
        """Read a pandas DataFrame from Parquet storage.

        Reads data from the specified location, optionally enforcing a schema
        and allowing filtering of data.

        Parameters
        ----------
        key : str
            Location/URL/path where data is persisted.
        filter_func : callable, optional
            Filter expression (see PyArrow Table.filter documentation).
        filters : list[tuple[str, str, list[str]]], optional
            Filter expression for read_table (see PyArrow parquet.read_table documentation).
        schema : pa.Schema, optional
            Data schema to enforce while reading.

        Returns
        -------
        pd.DataFrame
            The loaded DataFrame.
        """
        schema_msg_segment = "0" if schema is None else str(len(schema))
        logger.info(f"[get|in] ({key}, {filter_func}, {filters}, schema len:{schema_msg_segment})")
        logger.debug(f"[get|in] ({key}, {filter_func}, {filters}, {schema})")
        table = pq.read_table(key, filesystem=self.fs, filters=filters, schema=schema)
        if filter_func is not None:
            table = table.filter(filter_func)
        result = table.to_pandas(types_mapper=pandas_mapper)
        logger.info(f"[get|out] => {result.shape}")
        return result

    def delete(
        self,
        key: str,
        partition_field: str | None = None,
        partition_values: list[str] | None = None,
        kv_dict: dict[str, list[Any]] | None = None,
        schema: pa.Schema = None,
    ) -> None:
        """Delete partitions or data from Parquet storage.

        Removes partitions (full or partial), deletes specific values, or removes
        an entire dataset from Parquet storage.

        Parameters
        ----------
        key : str
            Location/URL/path where data is persisted.
        partition_field : str, optional
            Name of the partition field in the dataset.
        partition_values : list[str], optional
            Partition values to delete.
        kv_dict : dict[str, list[Any]], optional
            Key-value map defining fields and array of values for deletion filter.
        schema : pa.Schema, optional
            Data schema to enforce if reading is required.
        """
        schema_msg_segment = "0" if schema is None else str(len(schema))
        logger.info(
            f"[delete|in] ({key}, {partition_field}, {partition_values}, {kv_dict}, schema len:{schema_msg_segment})"
        )
        logger.debug(f"[delete|in] ({key}, {partition_field}, {partition_values}, {kv_dict}, {schema})")

        if partition_values is not None and partition_field is not None:
            self._remove_partitions(key, partition_field=partition_field, partition_values=partition_values)
        elif kv_dict is not None and partition_field is not None:
            table = pq.read_table(key, filesystem=self.fs, schema=schema)
            for k, v in kv_dict.items():
                filter_condition = ~pc.is_in(pc.field(k), pa.array(v))
                table = table.filter(filter_condition)
            self.delete(key, schema=schema)
            pq.write_to_dataset(
                table,
                root_path=key,
                partition_cols=[partition_field],
                existing_data_behavior="delete_matching",
                filesystem=self.fs,
                schema=schema,
            )
        else:
            self._rmdir(key)

        logger.info("[delete|out]")

    def save(
        self,
        df: pd.DataFrame,
        key: str,
        partition_field: str | None = None,
        append: bool = False,
        replace_partitions: bool = False,
        schema: Any = None,
    ) -> None:
        """Save a pandas DataFrame in Parquet format.

        Saves data to the specified location with optional partitioning,
        append, or replace behavior.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to be saved.
        key : str
            Location/URL/path where data should be persisted.
        partition_field : str, optional
            Name of the partition field in the dataset.
        append : bool, default False
            If True, data will be appended; otherwise will overwrite.
        replace_partitions : bool, default False
            If True, partitions will be replaced, deleting existing data in those partitions.
        schema : pa.Schema, optional
            Data schema to enforce while writing.

        Raises
        ------
        StoreException
            If both append and replace_partitions are True.
        """
        schema_msg_segment = "0" if schema is None else str(len(schema))
        logger.info(
            f"[save|in] ({df.shape}, {key}, {partition_field}, {append}, {replace_partitions}, schema len:{schema_msg_segment})"
        )
        logger.debug(f"[save|in] ({df}, {key}, {partition_field}, {append}, {replace_partitions}, {schema})")

        if schema is not None and isinstance(schema, pa.lib.Schema):
            # we will order the columns based on the schema
            columns = list(schema.names)
            df = df[columns]

        if replace_partitions and append:
            raise StoreException("cannot request for replace_partitions and append at the same time")

        if append:
            pq.write_to_dataset(
                pa.Table.from_pandas(df, preserve_index=False),
                root_path=key,
                partition_cols=[partition_field],
                filesystem=self.fs,
                schema=schema,
            )
        elif replace_partitions:
            partitions = df[partition_field].unique().tolist()
            self._remove_partitions(key, partition_field, partitions)
            pq.write_to_dataset(
                pa.Table.from_pandas(df, preserve_index=False),
                root_path=key,
                partition_cols=[partition_field],
                existing_data_behavior="delete_matching",
                filesystem=self.fs,
                schema=schema,
            )
        else:
            self.delete(key)
            pq.write_to_dataset(
                pa.Table.from_pandas(df, preserve_index=False),
                root_path=key,
                partition_cols=[partition_field],
                existing_data_behavior="delete_matching",
                filesystem=self.fs,
                schema=schema,
            )
        logger.info("[save|out]")

    def _remove_partitions(self, key: str, partition_field: str, partition_values: list[str]) -> None:
        """Remove specific partitions from the dataset.

        Parameters
        ----------
        key : str
            Root path of the dataset.
        partition_field : str
            Name of the partition field.
        partition_values : list[str]
            List of partition values to remove.
        """
        logger.debug(f"[_remove_partitions|in] ({key}, {partition_field}, {partition_values})")

        for partition_value in partition_values:
            partition_key = f"{partition_field}={partition_value}"
            partition_path = str(Path(key) / partition_key)
            self._rmdir(partition_path)

        logger.debug("[_remove_partitions|out]")

    def update(
        self,
        df: pd.DataFrame,
        key: str,
        key_fields: list[str],
        partition_field: str | None = None,
        schema: Any = None,
    ) -> None:
        """Update rows in a pandas DataFrame stored in Parquet format.

        Updates matching rows based on key fields by merging with existing data.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing updates.
        key : str
            Location/URL/path where data is persisted.
        key_fields : list[str]
            Primary fields used to match rows for updating.
        partition_field : str, optional
            Name of the partition field to enforce while saving.
        schema : pa.Schema, optional
            Data schema to enforce while reading and saving.
        """
        schema_msg_segment = "0" if schema is None else str(len(schema))
        logger.info(
            f"[update|in] ({df.shape}, {key}, {key_fields}, {partition_field}, schema len:{schema_msg_segment})"
        )
        logger.debug(f"[update|in] ({df}, {key}, {key_fields}, {partition_field}, {schema})")

        df0 = self.get(key, schema=schema)
        match = pd.merge(df0.reset_index(), df.reset_index(), on=key_fields)
        index_left = match["index_x"]
        index_right = match["index_y"]
        df0.iloc[index_left] = df.iloc[index_right]
        self.save(df0, key, partition_field=partition_field, schema=schema)

        logger.info("[update|out]")



"""Module for copying files between S3 buckets.

This module provides:
- S3FileCopy: class for copying objects/files from one S3 bucket to another
"""
import logging
from pathlib import Path
from typing import Any
from tgedr_dataops.commons.s3_connector import S3Connector

from tgedr_dataops_abs.source import Source, SourceException
from tgedr_dataops.commons.utils_fs import process_s3_path, resolve_s3_protocol


logger = logging.getLogger(__name__)


class S3FileCopy(Source, S3Connector):
    """class used to copy objects/files from an s3 bucket to another s3 bucket."""

    CONTEXT_KEY_SOURCE = "source"
    CONTEXT_KEY_TARGET = "target"
    CONTEXT_KEY_FILES = "files"
    CONTEXT_KEY_SUFFIX = "suffix"
    CONTEXT_KEY_PRESERVE_SOURCE_KEY = "preserve_source_key"

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        """Initialize the S3FileCopy source.

        Parameters
        ----------
        config : dict[str, Any], optional
            Configuration dictionary containing optional AWS credentials.
        """
        Source.__init__(self, config=config)
        S3Connector.__init__(self)

    def list(self, context: dict[str, Any] | None = None) -> list[str]:
        """List objects in an S3 bucket matching the given prefix.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' (source S3 URL) and optionally 'suffix' to filter results.

        Returns
        -------
        list[str]
            List of S3 URLs matching the criteria.

        Raises
        ------
        SourceException
            If the 'source' key is missing from the context.
        """
        logger.info(f"[list|in] ({context})")

        result: list[str] = []
        if self.CONTEXT_KEY_SOURCE not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_SOURCE}")

        s3_protocol: str = resolve_s3_protocol(context[self.CONTEXT_KEY_SOURCE])
        protocol = "" if s3_protocol is None else s3_protocol

        bucket, key = process_s3_path(context[self.CONTEXT_KEY_SOURCE])

        objs = self._client.list_objects_v2(Bucket=bucket, Prefix=key)
        result = [
            (protocol + bucket + "/" + entry["Key"]) for entry in objs["Contents"] if not (entry["Key"]).endswith("/")
        ]

        if self.CONTEXT_KEY_SUFFIX in context:
            suffix: str = context[self.CONTEXT_KEY_SUFFIX]
            result = [f for f in result if f.endswith(suffix)]

        logger.debug(f"[list|out] => {result}")
        logger.info(f"[list|out] => result len: {len(result)}")
        return result

    def get(self, context: dict[str, Any] | None = None) -> Any:
        """Copy a file from one S3 location to another.

        Parameters
        ----------
        context : dict[str, Any], optional
            Context dictionary containing 'source' (source S3 URL) and 'target' (target S3 URL).

        Returns
        -------
        Any
            The target S3 URL.

        Raises
        ------
        SourceException
            If source or target context is missing.
        """
        logger.info(f"[get|in] ({context})")

        result: list[str] = []
        if self.CONTEXT_KEY_FILES not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_FILES}")
        if self.CONTEXT_KEY_TARGET not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_TARGET}")

        preserve_source_key = False
        if self.CONTEXT_KEY_PRESERVE_SOURCE_KEY in context:
            preserve_source_key = str(context[self.CONTEXT_KEY_PRESERVE_SOURCE_KEY]).lower() in ["1", "true"]
        logger.info(f"[get] preserve_source_key: {preserve_source_key}")

        target_bucket, target_key = process_s3_path(context[self.CONTEXT_KEY_TARGET])
        s3_protocol: str = resolve_s3_protocol(context[self.CONTEXT_KEY_TARGET])
        protocol = "" if s3_protocol is None else s3_protocol

        files = context[self.CONTEXT_KEY_FILES]
        for file in files:
            src_bucket, src_key = process_s3_path(file)
            src = {"Bucket": src_bucket, "Key": src_key}

            target_key = target_key.rstrip("/") if target_key.endswith("/") else target_key
            if preserve_source_key:
                key = str(Path(target_key) / src_key)
            else:
                src_key_leaf = Path(src_key).name
                key = str(Path(target_key) / src_key_leaf)

            logger.info(
                f"[get] copying... src bucket: {src_bucket}   src key: {src_key}   target bucket: {target_bucket}   target key: {key}"
            )
            self._client.copy(src, target_bucket, key)

            result.append(str(Path(protocol) / target_bucket / key))

        logger.info(f"[get|out] => {result}")
        return result



"""Filesystem utility functions for file and path operations.

This module provides utilities for:
- Creating temporary files and directories
- Parsing and processing S3 URLs and paths
- Extracting URL protocols
- Generating file hashes
"""
import tempfile
import re
import hashlib
from pathlib import Path


def temp_dir(root: str | None = None, suffix: str | None = None, prefix: str | None = None) -> str:
    """Create a temporary directory and return its path.

    Parameters
    ----------
    root : str | None
        Directory where the temporary directory will be created.
    suffix : str | None
        Suffix for the temporary directory name.
    prefix : str | None
        Prefix for the temporary directory name.

    Returns
    -------
    str
        Path to the created temporary directory.
    """
    return tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=root)


def temp_file(
    root: str | None = None, suffix: str | None = None, prefix: str | None = None,
    discard_handle: bool = True
) -> str | tuple[int, str]:
    """Create a temporary file and return its path or handle and path.

    Parameters
    ----------
    root : str | None
        Directory where the temporary file will be created.
    suffix : str | None
        Suffix for the temporary file name.
    prefix : str | None
        Prefix for the temporary file name.
    discard_handle : bool
        If True, return only the file path. If False, return tuple of (handle, path).

    Returns
    -------
    str | tuple[int, str]
        File path if discard_handle is True, otherwise tuple of (file handle, file path).
    """
    h, f = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=root)
    if discard_handle:
        return f
    return (h, f)


def resolve_url_protocol(url: str) -> str:
    """Extract the protocol from a URL.

    Parameters
    ----------
    url : str
        The URL to extract the protocol from.

    Returns
    -------
    str
        The protocol (e.g., 'http://', 'https://') or None if no protocol is found.
    """
    result = None
    group_match = re.search("(.*://).*", url)
    if group_match is not None:
        result = group_match.group(1)
    return result


def resolve_s3_protocol(url: str) -> str:
    """Extract the S3 protocol from a URL.

    Parameters
    ----------
    url : str
        The URL to extract the S3 protocol from.

    Returns
    -------
    str
        The S3 protocol (e.g., 's3://', 's3a://') or None if no S3 protocol is found.
    """
    result = None
    group_match = re.search("(s3[a]?://).*", url)
    if group_match is not None:
        result = group_match.group(1)
    return result


def remove_s3_protocol(url: str) -> str:
    """Remove the S3 protocol prefix from a URL.

    Parameters
    ----------
    url : str
        The S3 URL to remove the protocol from.

    Returns
    -------
    str
        The URL without the S3 protocol prefix (s3:// or s3a://).
    """
    if url.startswith("s3://"):
        result = url[5:]
    elif url.startswith("s3a://"):
        result = url[6:]
    else:
        result = url
    return result


def process_s3_path(path: str) -> tuple[str, str]:
    """Extract bucket and key from an S3 path.

    Parameters
    ----------
    path : str
        The S3 path to process (with or without protocol).

    Returns
    -------
    tuple[str, str]
        A tuple containing (bucket, key).
    """
    no_protocol_path = remove_s3_protocol(path)
    path_elements = no_protocol_path.split("/")
    bucket = path_elements[0]
    key = "/".join(path_elements[1:])
    return (bucket, key)


def process_s3_url(url: str) -> tuple[str, str, str]:
    """Extract protocol, bucket, and key from an S3 URL.

    Parameters
    ----------
    url : str
        The S3 URL to process.

    Returns
    -------
    tuple[str, str, str]
        A tuple containing (protocol, bucket, key).
    """
    protocol = resolve_s3_protocol(url)
    no_protocol_url = remove_s3_protocol(url)
    path_elements = no_protocol_url.split("/")
    bucket = path_elements[0]
    key = "/".join(path_elements[1:])
    return ("" if protocol is None else protocol, bucket, key)


def hash_file(filepath: str, hash_func=hashlib.sha256) -> str:  # noqa: ANN001
    """Generate a hash for a file.

    Args:
        filepath (str): The path to the file.
        hash_func: A hashlib hash function, e.g., hashlib.md5().

    Returns:
        str: The hexadecimal hash string of the file.
    """
    # Initialize the hash object
    hasher = hash_func()

    # Open the file in binary read mode
    with Path(filepath).open("rb") as file:
        # Read the file in chunks to avoid using too much memory
        chunk_size = 8192
        while chunk := file.read(chunk_size):
            hasher.update(chunk)

    # Return the hexadecimal digest of the hash
    return hasher.hexdigest()
